import logging
from typing import Optional

from meross_iot.model.enums import Namespace, SprayMode

_LOGGER = logging.getLogger(__name__)


class SprayMixin(object):
    _execute_command: callable
    check_full_update_done: callable
    #async_handle_update: Callable[[Namespace, dict], Awaitable]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

        # Dictionary keeping the status for every channel
        self._channel_spray_status = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.CONTROL_SPRAY:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace {namespace}")
            payload = data.get('spray')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'spray' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # Update the status of every channel that has been reported in this push
                # notification.
                for c in payload:
                    channel = c['channel']
                    strmode = c['mode']
                    mode = SprayMode(strmode)
                    self._channel_spray_status[channel] = mode

                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    def get_current_mode(self, channel: int = 0, *args, **kwargs) -> Optional[SprayMode]:
        self.check_full_update_done()
        return self._channel_spray_status.get(channel)

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            spray_data = data.get('all', {}).get('digest', {}).get('spray', [])
            for c in spray_data:
                channel = c['channel']
                strmode = c['mode']
                mode = SprayMode(strmode)
                self._channel_spray_status[channel] = mode
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    async def async_set_mode(self, mode: SprayMode, channel: int = 0, timeout: Optional[float] = None, *args, **kwargs) -> None:
        payload = {'spray': {'channel': channel, 'mode': mode.value}}
        await self._execute_command(method='SET',
                                    namespace=Namespace.CONTROL_SPRAY,
                                    payload=payload,
                                    timeout=timeout)
        # Immediately update local state
        self._channel_spray_status[channel] = mode



import logging
import re

_LOGGER = logging.getLogger(__name__)


camel_pat = re.compile(r'([A-Z])')
under_pat = re.compile(r'_([a-z])')


def _camel_to_underscore(key):
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), key)


def _underscore_to_camel(key):
    return under_pat.sub(lambda x: x.group(1).upper(), key)


class BaseDictPayload(object):
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def from_dict(cls, json_dict: dict):
        # Transform the camel-case notation into a more pythonian case
        new_dict = {_camel_to_underscore(key): value for (key, value) in json_dict.items()}
        obj = cls(**new_dict)
        return obj

    def to_dict(self) -> dict:
        res = {}
        for k, v in vars(self).items():
            new_key = _underscore_to_camel(k)
            res[new_key] = v
        return res



import logging
from typing import Optional, List

from meross_iot.controller.device import ChannelInfo
from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)


class GarageOpenerMixin:
    _channels: List[ChannelInfo]
    _execute_command: callable
    check_full_update_done: callable
    uuid: str

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._door_open_state_by_channel = {}
        self._door_config_state_by_channel = {}

        # Initialize the state attributes
        for c in self._channels:
            self._door_open_state_by_channel[c.index] = None
            self._door_config_state_by_channel[c.index] = None

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.GARAGE_DOOR_STATE:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('state')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'state' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The door opener state push notification contains an object for every channel handled by the
                # device.
                for door in payload:
                    channel_index = door['channel']
                    state = door['open'] == 1
                    self._door_open_state_by_channel[channel_index] = state
                    locally_handled = True
        elif namespace == Namespace.GARAGE_DOOR_MULTIPLECONFIG:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('config')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'config' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The door opener state push notification contains an object for every channel handled by the
                # device
                for door in payload:
                    channel_index = door['channel']
                    self._door_config_state_by_channel[channel_index] = door
                    locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            doors_data = data.get('all', {}).get('digest', {}).get('garageDoor', [])
            for door in doors_data:
                channel_index = door['channel']
                state = door['open'] == 1
                self._door_open_state_by_channel[channel_index] = state
            locally_handled = True

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    async def async_open(self, channel: Optional[int] = None, *args, **kwargs) -> None:
        """
        Operates the door: sends the open command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(state=True, channel=channel, *args, **kwargs)

    async def async_close(self, channel: Optional[int] = None, *args, **kwargs) -> None:
        """
        Operates the door: sends the close command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(state=False, channel=channel, *args, **kwargs)

    async def _async_operate(self,
                             state: bool,
                             channel: Optional[int] = None,
                             timeout: Optional[float] = None,
                             *args, **kwargs) -> None:
        target_channel = self._get_default_channel_index(channel)
        payload = {"state": {"channel": target_channel, "open": 1 if state else 0, "uuid": self.uuid}}
        await self._execute_command(method="SET",
                                    namespace=Namespace.GARAGE_DOOR_STATE,
                                    payload=payload,
                                    timeout=timeout)

    def get_is_open(self, channel: Optional[int] = None, *args, **kwargs) -> Optional[bool]:
        """
        The current door-open status. Returns True if the given door is open, False otherwise.
        :param channel: channel of which status is needed

        :return: False if the door is closed, True otherwise
        """
        self.check_full_update_done()
        target_channel = self._get_default_channel_index(channel)
        return self._door_open_state_by_channel.get(target_channel, None)

    def _get_default_channel_index(self, channel: Optional[int]) -> int:
        if channel is not None:
            return channel

        if len(self._channels)<2:
            return 0
        else:
            return 1


import logging
from datetime import datetime
from typing import List, Optional

from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)


_DATE_FORMAT = '%Y-%m-%d'


class ConsumptionXMixin(object):
    _execute_command: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

    async def async_get_daily_power_consumption(self,
                                                channel=0,
                                                timeout: Optional[float] = None,
                                                *args, **kwargs) -> List[dict]:
        """
        Returns the power consumption registered by this device.

        :param channel: channel to read data from

        :return: the historical consumption data
        """
        # TODO: returning a nice PowerConsumtpionReport object rather than a list of dict?
        result = await self._execute_command(method="GET",
                                             namespace=Namespace.CONTROL_CONSUMPTIONX,
                                             payload={'channel': channel},
                                             timeout=timeout)
        data = result.get('consumptionx')

        # Parse the json data into nice-python native objects
        res = [{
                'date': datetime.strptime(x.get('date'), _DATE_FORMAT),
                'total_consumption_kwh': float(x.get('value'))/1000
        } for x in data]

        return res


class ConsumptionMixin(object):
    _execute_command: callable

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

    async def async_get_daily_power_consumption(self,
                                                channel=0,
                                                timeout: Optional[float] = None,
                                                *args, **kwargs) -> List[dict]:
        """
        Returns the power consumption registered by this device.

        :param channel: channel to read data from

        :return: the historical consumption data
        """
        # TODO: returning a nice PowerConsumtpionReport object rather than a list of dict?
        result = await self._execute_command(method="GET",
                                             namespace=Namespace.CONTROL_CONSUMPTION,
                                             payload={'channel': channel},
                                             timeout=timeout)
        data = result.get('consumption')

        # Parse the json data into nice-python native objects
        res = [{
                'date': datetime.strptime(x.get('date'), _DATE_FORMAT),
                'total_consumption_kwh': float(x.get('value'))/1000
        } for x in data]

        return res


import asyncio
import datetime
import getpass
import json
import logging
import os
import sys
import tempfile
from concurrent.futures import ThreadPoolExecutor
from hashlib import md5
from os import path, environ
from typing import List, Dict, Tuple
from zipfile import ZipFile


from meross_iot.http_api import MerossHttpClient
from meross_iot.manager import MerossManager, TransportMode
from meross_iot.model.credentials import MerossCloudCreds
from meross_iot.model.enums import Namespace, OnlineStatus
from meross_iot.model.http.device import HttpDeviceInfo
from utilities.meross_fake_app import AppSniffer
from utilities.meross_fake_device import FakeDeviceSniffer
from urllib.parse import urlparse

SNIFF_LOG_FILE = 'sniff.log'
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# Configure logging
logging.basicConfig(level=logging.DEBUG)
l = logging.getLogger()
l.handlers.clear()

# Debug file handler
file_handler = logging.FileHandler(filename='sniff.log', mode='w+')
file_handler.setFormatter(formatter)
file_handler.setLevel(logging.DEBUG)
l.addHandler(file_handler)

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.CRITICAL)
l.addHandler(stream_handler)


def _print_welcom_message():
    print("Welcome to the Sniffer utility. This python script will gather some useful information about your "
          "Meross devices. All the collected information will be zipped into a zip archive. "
          "You could share such zip file with the developers to help them add support for your device. "
          "Although this utility won't collect your email/password, we recommend you to change "
          "your Meross account password to a temporary one before using this software. Once you are done, "
          "you can restore back your original password. By doing so, you are 100% sure you are not leaking any "
          "real password to the developers.\n\n"
          "ATTENTION: The sniffer does not support MFA login. Please disable MFA if willing to use this script.\n\n")


async def _async_gather_http_client() -> MerossHttpClient:
    api_base_url = environ.get("MEROSS_API_URL", "https://iotx-us.meross.com")
    if api_base_url is None:
        api_base_url = input("Please specify the meross API base url. It should be 'https://iotx-us.meross.com' or 'https://iotx-eu.meross.com' or 'https://iotx-ap.meross.com'. "
                             "Leave blank if unsure: ")
        api_base_url = api_base_url.strip()
        if api_base_url is None:
            api_base_url = "https://iotx-us.meross.com"
            print(f"Assuming API Base URL: {api_base_url}")
        try:
            urlparse(api_base_url)
        except:
            print("Invalid API/URL provided.")
            exit(2)
    email = environ.get("MEROSS_EMAIL")
    if email is None:
        email = input("Please specify your meross email: ")
        email = email.strip()

    password = environ.get("MEROSS_PASSWORD")
    if password is None:
        password = getpass.getpass(prompt='Please specify your Meross password: ', stream=sys.stdout).strip()
    try:
        return await MerossHttpClient.async_from_user_password(api_base_url=api_base_url, email=email, password=password)
    except Exception:
        print("An error occurred while gathering MerossAPI Client. Make sure your email-password credentials are valid.")
        exit(1)


async def _async_print_device_list(client: MerossHttpClient) -> List[HttpDeviceInfo]:
    # Gather HTTP devices
    try:
        print("# Collecting devices via HTTP api...")
        devices = await client.async_list_devices()
        for i, d in enumerate(devices):
            print(f"[{i}] - {d.dev_name} ({d.device_type}) - {d.online_status.name}")
        return devices
    except:
        print("An error occurred while retrieving Meross devices.")
        exit(1)


async def _async_select_device(devices: List[HttpDeviceInfo]) -> HttpDeviceInfo:
    print("Please note that some devices rely on a HUB. Those devices won't be listed: in such cases, you need"
          "to select the corresponding HUB device.")
    while True:
        selection = input("Select the device you want to study (numeric index): ")
        selection = int(selection.strip())
        selected_device = devices[selection]
        if selected_device is not None:
            break
    print(f"You have selected {selected_device.dev_name}.")
    if selected_device.online_status != OnlineStatus.ONLINE:
        print("!! WARNING !! You selected a device that has not been reported as online. ")
    return selected_device


async def _async_collect_device_base_data(client: MerossHttpClient, selected_device: HttpDeviceInfo) -> Tuple[Dict, MerossManager]:
    print("Collecting devices info...")
    manager = MerossManager(http_client=client)
    await manager.async_init()
    mqtt_host = selected_device.get_mqtt_host()
    mqtt_port = selected_device.get_mqtt_port()

    # Manually get device abilities
    print(f"Collecting {Namespace.SYSTEM_ALL} info...")
    system_data = await manager.async_execute_cmd(destination_device_uuid=selected_device.uuid,
                                                   method="GET",
                                                   namespace=Namespace.SYSTEM_ALL,
                                                   payload={},
                                                   mqtt_hostname=mqtt_host,
                                                   mqtt_port=mqtt_port)
    l.info(f"Sysdata for {selected_device.dev_name} ({selected_device.uuid}): {system_data}")

    try:
        print(f"Collecting {Namespace.SYSTEM_ABILITY} info...")
        response_abilities = await manager.async_execute_cmd(destination_device_uuid=selected_device.uuid,
                                                             method="GET",
                                                             namespace=Namespace.SYSTEM_ABILITY,
                                                             payload={},
                                                             mqtt_hostname=mqtt_host,
                                                             mqtt_port=mqtt_port
                                                             )

        l.info(f"Abilities for {selected_device.dev_name} ({selected_device.uuid}): {response_abilities}")
    except:
        l.exception(f"Could not collect sysdata/abilities for {selected_device.uuid}")

    return system_data, manager


async def _async_start_fake_device_sniffer(cloud_credentials: MerossCloudCreds, selected_device: HttpDeviceInfo, mac_address: str) -> FakeDeviceSniffer:
    print("Starting the fake device emulator...")
    fake_device_sniffer = FakeDeviceSniffer(uuid=selected_device.uuid, mac_address=mac_address,
                                            meross_user_id=cloud_credentials.user_id,
                                            meross_cloud_key=cloud_credentials.key,
                                            mqtt_host=selected_device.get_mqtt_host(),
                                            mqtt_port=selected_device.get_mqtt_port(),
                                            logger=l)
    await fake_device_sniffer.async_start(timeout=5.0)
    return fake_device_sniffer


def _start_app_sniffer(cloud_credentials: MerossCloudCreds, selected_device: HttpDeviceInfo) -> AppSniffer:
    md5_hash = md5()
    clearpwd = "%s%s" % (cloud_credentials.user_id, cloud_credentials.key)
    md5_hash.update(clearpwd.encode("utf8"))
    hashed_password = md5_hash.hexdigest()
    app_sniffer = AppSniffer(
        l,
        cloud_credentials.user_id,
        hashed_password,
        selected_device.uuid,
        ca_cert=None,
        mqtt_host=selected_device.get_mqtt_host(),
        mqtt_port=selected_device.get_mqtt_port()
    )

    print("Starting the app-simulator sniffer...")
    app_sniffer.start()
    return app_sniffer


async def _async_sniff(device_sniffer: FakeDeviceSniffer, zip_obj: ZipFile, meross_manager: MerossManager, selected_device: HttpDeviceInfo):
    print("--------------------------------------\n"
          "PHASE 1: intercepting device commands.\n"
          "--------------------------------------\n"
          "The goal of this phase is to 'intercept' the commands that the APP sends to the device, pretending to"
          "be that specific device. In other words, we will 'spoof' the device connection so that the commands "
          "sent by the official APP are instead routed to this utility.\n\n"
          "To do so, you need to disconnect the real device from the 'power outlet', so that it gets cut out from "
          "the Meross cloud, then issue the command you want to inspect from the Meross App. This command will FAIL,"
          "as the message will not reach the real Meross Device, but this utility instead. When the command is "
          "received, it will be printed out.\n"
          "If you need to test multiple states/commands, you can always re-attach the power to the real Meross device,"
          "perform adjust its original state and then cut off the power and issue the command you want to get info.\n"
          "To simplify this process, the utility will guide you throughout the process.\n"
          "\nXXXXXXXXXXXXXX ATTENTION XXXXXXXXXXXXXX\n"
          "XXX MAKE SURE THE APP IS NOT CONNECTED TO THE SAME WIFI OF THE MEROSS DEVICE."
          "XXX If both the APP and the Meross device are on the same WiFi network, their communication will not "
          "XXX be intercepted. So, it's BETTER to disable the WiFi completely on the smartphone and rely on 3g/4g/5g "
          "XXX connectivity. That will make sure the APP uses the MQTT channel to talk to the device, and so the "
          "XXX communication will be intercepted by this sniffer."
          "\nXXXXXXXXXXXXXX ATTENTION XXXXXXXXXXXXXX\n\n"
          )

    messages = set()
    executor = ThreadPoolExecutor(2)

    with tempfile.NamedTemporaryFile(mode="w+t", encoding="utf8", prefix=f"sniffed_commands", suffix=".txt", delete=False) as f:
        residual_timeout = 30.0
        print(f"Waiting up to 30 seconds for messages...")
        while True:
            try:
                before = datetime.datetime.now().timestamp()
                raw_message, namespace, method, payload = await asyncio.wait_for(device_sniffer.async_wait_for_message(), timeout=residual_timeout)
                after = datetime.datetime.now().timestamp()
                key = f"{method}{namespace}{payload}"
                if key in messages:
                    # This message type was already intercepted
                    residual_timeout -= (after - before)
                    if residual_timeout > 0:
                        continue
                    else:
                        raise TimeoutError()
            except asyncio.exceptions.TimeoutError:
                resp = input("Timeout: no message received. Do you want to wait more? [y/N]:").strip().lower()
                if resp in ('y','yes'):
                    residual_timeout = 30
                    print(f"Waiting up to 30 seconds for messages...")
                    continue
                else:
                    break

            messages.add(key)
            print(f"New message type received: {method} {namespace}: {payload}")
            note = input("Describe the command you issued on the APP which caused this message. If unsure, just type \"?\" or \"IDK\": ")
            f.write(f"\n\nCommand Description: {note}\n\n")
            f.write(f"Request: \n")
            f.write(raw_message.payload.decode('utf8'))
            f.flush()

            # Attempt now to retrieve a response simulating that command to the real device
            print("Replicating the same message to the device in order to catch the real response...")
            try:
                response = await meross_manager.async_execute_cmd(
                    mqtt_hostname=selected_device.get_mqtt_host(),
                    mqtt_port=selected_device.get_mqtt_port(),
                    destination_device_uuid=selected_device.uuid,
                    method=method,
                    namespace=namespace,
                    payload=payload,
                    timeout=5.0,
                    override_transport_mode=TransportMode.MQTT_ONLY
                )
                f.write(f"\nResponse: \n")
                f.write(json.dumps(response))
            except Exception as e:
                l.exception("An error occurred while collecting a response from the device simulating the latest message. No response will be catched.")
                print("Skipping response collection.")
            print(f"Waiting up to 30 seconds for messages...")
        zip_obj.write(f.name, f"sniffed_commands.txt")


async def _main():
    zip_obj = ZipFile('data.zip', 'w')
    _print_welcom_message()
    client = await _async_gather_http_client()
    manager = None
    try:
        devices = await _async_print_device_list(client)
        selected_device = await _async_select_device(devices)

        # Log/Collect device data device data
        system_data, manager = await _async_collect_device_base_data(client, selected_device)
        device_mac_address = system_data["all"]["system"]["hardware"]["macAddress"]

        # Start the device sniffer
        device_sniffer = await _async_start_fake_device_sniffer(client.cloud_credentials, selected_device, device_mac_address)

        # Start the app sniffer
        app_sniffer = _start_app_sniffer(client.cloud_credentials, selected_device)

        # Start simulation and sniffing (phase 1)
        await _async_sniff(device_sniffer, zip_obj, manager, selected_device)
        await device_sniffer.async_stop()
        # Start the PUSH notification catching (phase 2)
        app_sniffer.stop()

    finally:
        if client is not None:
            await client.async_logout()
        if manager is not None:
            manager.close()

        print("Collecting logs...")
        zip_obj.write(SNIFF_LOG_FILE, 'sniff_log.txt')
        zip_obj.close()

        print("A zipfile has been created containing the logs collected during this execution. "
              "It is located in {path}.".format(path=path.abspath(zip_obj.filename)))
        print("Thanks for helping the Meross community!")



def main():
    # On Windows + Python 3.8, you should uncomment the following
    if os.name == 'nt':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_main())
    loop.stop()


if __name__ == '__main__':
    main()



import os
from ctypes import Union
from typing import Optional

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.consumption import ConsumptionXMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus
from tests import async_get_client

# Fix event loop for Windows env
if os.name == 'nt':
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestConsumptionX(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        toggle_devices = self.meross_manager.find_devices(device_class=ConsumptionXMixin, online_status=OnlineStatus.ONLINE)

        if len(toggle_devices) < 1:
            self.test_device = None
        else:
            self.test_device = toggle_devices[0]

    @unittest_run_loop
    async def test_consumptionx_local_state(self):
        if self.test_device is None:
            self.skipTest("No ConsumptionX device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")
        r = await self.test_device.async_get_daily_power_consumption()
        self.assertGreater(len(r), 1)

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import os
import random
from random import randint
from typing import List, TypeVar

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.thermostat import ThermostatModeMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus, ThermostatMode
from tests import async_get_client

if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio

Mts200Type = TypeVar('Mts200Type', BaseDevice, ThermostatModeMixin)


class TestThermostat(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        await self.meross_manager.async_device_discovery()
        self.test_devices: List[Mts200Type] = self.meross_manager.find_devices(device_class=ThermostatModeMixin, online_status=OnlineStatus.ONLINE)

    @unittest_run_loop
    async def test_on_off(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        toggled_state = not state.is_on
        await dev.async_set_thermostat_config(on_not_off=toggled_state)
        self.assertEqual(dev.get_thermostat_state().is_on, toggled_state)

    @unittest_run_loop
    async def test_ambient_temperature(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        for dev in self.test_devices:
            print(f"Testing device {dev.name}")
            res = await dev.async_update()
            temperature = dev.get_thermostat_state().current_temperature_celsius
            print(f"Thermostat {dev.name} reports ambiente temperature of {temperature} °C")
            self.assertIsInstance(temperature, float)

    @unittest_run_loop
    async def test_mode(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        self.assertIsNotNone(state.mode)
        modes = set(ThermostatMode)
        modes.remove(state.mode)
        modes = list(modes)
        index = randint(0, len(modes)-1)
        target_mode = modes[index]

        await dev.async_set_thermostat_config(mode=target_mode)
        self.assertEqual(target_mode, dev.get_thermostat_state().mode)

    def _align_temp(self, temperature:float) -> float:
        # Round temp value to 0.5
        quotient = temperature/0.5
        quotient = round(quotient)
        final_temp = quotient*0.5
        return final_temp

    @unittest_run_loop
    async def test_eco_temp(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        min_temp = state.min_temperature_celsius
        max_temp = state.max_temperature_celsius
        target_temp = min_temp + random.random()* (max_temp-min_temp)
        target_temp = self._align_temp(target_temp)
        # Set eco temp mode
        await dev.async_set_thermostat_config(eco_temperature_celsius=target_temp)
        # Ensure eco temp mode has been updated
        self.assertEqual(dev.get_thermostat_state().eco_temperature_celsius, target_temp)
        # Set eco mode
        await dev.async_set_thermostat_config(mode=ThermostatMode.ECONOMY)
        self.assertEqual(dev.get_thermostat_state().mode, ThermostatMode.ECONOMY)

    @unittest_run_loop
    async def test_cool_temp(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        min_temp = state.min_temperature_celsius
        max_temp = state.max_temperature_celsius
        target_temp = min_temp + random.random() * (max_temp - min_temp)
        target_temp = self._align_temp(target_temp)
        # Set eco temp mode
        await dev.async_set_thermostat_config(cool_temperature_celsius=target_temp)
        # Ensure eco temp mode has been updated
        self.assertEqual(dev.get_thermostat_state().cool_temperature_celsius, target_temp)
        # Set eco mode
        await dev.async_set_thermostat_config(mode=ThermostatMode.COOL)
        self.assertEqual(dev.get_thermostat_state().mode, ThermostatMode.COOL)

    @unittest_run_loop
    async def test_heat_temp(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        min_temp = state.min_temperature_celsius
        max_temp = state.max_temperature_celsius
        target_temp = min_temp + random.random() * (max_temp - min_temp)
        target_temp = self._align_temp(target_temp)
        # Set eco temp mode
        await dev.async_set_thermostat_config(heat_temperature_celsius=target_temp)
        # Ensure eco temp mode has been updated
        self.assertEqual(dev.get_thermostat_state().heat_temperature_celsius, target_temp)
        # Set eco mode
        await dev.async_set_thermostat_config(mode=ThermostatMode.HEAT)
        self.assertEqual(dev.get_thermostat_state().mode, ThermostatMode.HEAT)

    @unittest_run_loop
    async def test_manual_temp(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        state = dev.get_thermostat_state()
        min_temp = state.min_temperature_celsius
        max_temp = state.max_temperature_celsius
        target_temp = min_temp + random.random() * (max_temp - min_temp)
        target_temp = self._align_temp(target_temp)
        # Set eco temp mode
        await dev.async_set_thermostat_config(manual_temperature_celsius=target_temp)
        # Ensure eco temp mode has been updated
        self.assertEqual(dev.get_thermostat_state().manual_temperature_celsius, target_temp)
        # Set eco mode
        await dev.async_set_thermostat_config(mode=ThermostatMode.MANUAL)
        self.assertEqual(dev.get_thermostat_state().mode, ThermostatMode.MANUAL)

    @unittest_run_loop
    async def test_onoff(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        for dev in self.test_devices:
            print(f"Testing device {dev.name}")
            res = await dev.async_update()

            # Turn off
            await dev.async_set_thermostat_config(on_not_off=False)
            self.assertFalse(dev.get_thermostat_state().is_on)
            await asyncio.sleep(1)
            # Turn on
            await dev.async_set_thermostat_config(on_not_off=True)
            self.assertTrue(dev.get_thermostat_state().is_on)
            await asyncio.sleep(1)
            # Turn off
            await dev.async_set_thermostat_config(on_not_off=False)
            self.assertFalse(dev.get_thermostat_state().is_on)
            await asyncio.sleep(1)

    #
    # @unittest_run_loop
    # async def test_push_notification(self):
    #     if len(self.test_devices) < 1:
    #         self.skipTest("No valve device has been found to run this test.")
    #
    #     dev1 = self.test_devices[0]
    #     print(f"Testing device {dev1.name}")
    #     # Turn it on
    #     await dev1.async_turn_on()
    #
    #     # Create a new manager
    #     new_meross_client, requires_logout = await async_get_client()
    #     m = None
    #     try:
    #         # Retrieve the same device with another manager
    #         m = MerossManager(http_client=new_meross_client)
    #         await m.async_init()
    #         await m.async_device_discovery()
    #         devs = m.find_devices(internal_ids=(dev1.internal_id,))
    #         dev = devs[0]
    #
    #         await dev.async_update()
    #         await dev1.async_update()
    #
    #         # Set target temperature to a random state
    #         target = randint(dev.min_supported_temperature, dev.max_supported_temperature)
    #         print(f"TARGET = {target}...")
    #         await dev1.async_set_target_temperature(temperature=target)
    #
    #         # The manager that issues the command would immediately update the local state, so we can check
    #         # its update as soon as the command is issued.
    #         dev1_target_temp = dev1.target_temperature
    #         print(f"DEV1 = {dev1_target_temp}...")
    #         self.assertEqual(dev1_target_temp, target)
    #
    #         # Wait a bit: give time for the push notification to get received on the other manager...
    #         await asyncio.sleep(5)
    #         # Make sure the other manager has received the push notification event
    #         dev_target_temp = dev.target_temperature
    #         print(f"DEV = {dev_target_temp}...")
    #         self.assertEqual(dev_target_temp, target)
    #
    #     finally:
    #         if m is not None:
    #             m.close()
    #         if requires_logout:
    #             await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import json
from datetime import datetime
from typing import Union

from meross_iot.model.constants import DEFAULT_MQTT_HOST


class MerossCloudCreds(object):
    """
    Utility class that holds Meross Cloud credentials for a given user.
    """
    def __init__(self, token: str, key: str, user_id: str, user_email: str, issued_on: Union[datetime, str], domain: str, mqtt_domain :str):
        self.token = token
        self.key = key
        self.user_id = user_id
        self.user_email = user_email
        self.domain = domain
        self.mqtt_domain = mqtt_domain,

        if isinstance(issued_on, datetime):
            self.issued_on = issued_on
        else:
            self.issued_on = datetime.fromisoformat(issued_on)

    def to_json(self) -> str:
        """
        Serializes the current object to a json string
        :return:
        """
        return json.dumps({
            "token": self.token,
            "key": self.key,
            "user_id": self.user_id,
            "user_email": self.user_email,
            "domain": self.domain,
            "mqtt_domain": self.mqtt_domain,
            "issued_on": self.issued_on.isoformat()
        })

    @classmethod
    def from_json(cls, json_string: str):
        """
        Builds a MerossCloudCreds instance from a previously serialized json string
        :param json_string:
        :return:
        """
        data = json.loads(json_string)
        # The old version of this object did not store the domain and mqttDomain.
        # We add in here the logic we need to ensure backward compatibility.
        if 'mqtt_domain' not in data:
            data['mqtt_domain'] = DEFAULT_MQTT_HOST
        if 'domain' not in data:
            data['domain'] = "http://iot.meross.com"
        res = MerossCloudCreds(**data)
        return res

    def __repr__(self):
        return self.to_json()



import os
import random
from typing import Union, Dict

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.roller_shutter import RollerShutterTimerMixin
from meross_iot.manager import MerossManager
from meross_iot.model.enums import OnlineStatus, Namespace, RollerShutterState
from tests import async_get_client

if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


DEFAULT_OPEN_TIMER = 15
DEFAULT_CLOSE_TIMER = 15


class TestRollerShutter(AioHTTPTestCase):
    test_device: Union[RollerShutterTimerMixin, BaseDevice, None]

    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()
        roller_devices = self.meross_manager.find_devices(device_class=RollerShutterTimerMixin,
                                                          online_status=OnlineStatus.ONLINE)

        if len(roller_devices) < 1:
            self.test_device = None
        else:
            self.test_device = roller_devices[0]

    @unittest_run_loop
    async def test_open(self):
        if self.test_device is None:
            self.skipTest("No RollerShutter device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        # Update its status
        await self.test_device.async_update()

        # Set timers
        await self.test_device.async_set_config(open_timer_seconds=DEFAULT_OPEN_TIMER,
                                                close_timer_seconds=DEFAULT_CLOSE_TIMER)

        state_opening = asyncio.Event()
        state_closing = asyncio.Event()
        state_idle = asyncio.Event()
        position_opened = asyncio.Event()
        position_closed = asyncio.Event()

        async def coro(namespace: Namespace, data: Dict, device_internal_id: str):
            if namespace == Namespace.ROLLER_SHUTTER_STATE:
                states = data.get('state')
                # Filter by channel
                state = next(filter(lambda s: s.get('channel') == 0, states)).get('state')
                if state == RollerShutterState.OPENING.value:
                    state_opening.set()
                elif state == RollerShutterState.CLOSING.value:
                    state_closing.set()
                elif state == RollerShutterState.IDLE.value:
                    state_idle.set()
            if namespace == Namespace.ROLLER_SHUTTER_POSITION:
                positions = data.get('position')
                # Filter by channel
                position = next(filter(lambda s: s.get('channel') == 0, positions)).get('position')
                if position == 100:
                    position_opened.set()
                elif position == 0:
                    position_closed.set()

        self.test_device.register_push_notification_handler_coroutine(coro)

        # Trigger the opening
        print("Sending opening command to Roller Shutter")
        await self.test_device.async_open()
        print("Waiting for state to become OPENING...")
        await asyncio.wait_for(state_opening.wait(), timeout=30.0)
        print("Waiting for state to become IDLE...")
        await asyncio.wait_for(state_idle.wait(), timeout=60.0)
        print("Waiting for position to become 100...")
        await asyncio.wait_for(position_opened.wait(), timeout=30.0)
        print("DONE!")

        self.test_device.unregister_push_notification_handler_coroutine(coro)

    @unittest_run_loop
    async def test_close(self):
        if self.test_device is None:
            self.skipTest("No RollerShutter device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        # Set timers
        await self.test_device.async_set_config(open_timer_seconds=DEFAULT_OPEN_TIMER, close_timer_seconds=DEFAULT_CLOSE_TIMER)

        # Update its status
        await self.test_device.async_update()

        state_opening = asyncio.Event()
        state_closing = asyncio.Event()
        state_idle = asyncio.Event()
        position_opened = asyncio.Event()
        position_closed = asyncio.Event()

        async def coro(namespace: Namespace, data: Dict, device_internal_id: str):
            if namespace == Namespace.ROLLER_SHUTTER_STATE:
                states = data.get('state')
                # Filter by channel
                state = next(filter(lambda s: s.get('channel') == 0, states)).get('state')
                if state == RollerShutterState.OPENING.value:
                    state_opening.set()
                elif state == RollerShutterState.CLOSING.value:
                    state_closing.set()
                elif state == RollerShutterState.IDLE.value:
                    state_idle.set()
            if namespace == Namespace.ROLLER_SHUTTER_POSITION:
                positions = data.get('position')
                # Filter by channel
                position = next(filter(lambda s: s.get('channel') == 0, positions)).get('position')
                if position == 100:
                    position_opened.set()
                elif position == 0:
                    position_closed.set()

        self.test_device.register_push_notification_handler_coroutine(coro)

        # Trigger the closing
        print("Sending closing command to Roller Shutter")
        await self.test_device.async_close()
        print("Waiting for state to become CLOSING...")
        await asyncio.wait_for(state_closing.wait(), timeout=30.0)
        print("Waiting for state to become IDLE...")
        await asyncio.wait_for(state_idle.wait(), timeout=60.0)
        print("Waiting for position to become 0...")
        await asyncio.wait_for(position_closed.wait(), timeout=30.0)
        print("DONE!")

        self.test_device.unregister_push_notification_handler_coroutine(coro)

    @unittest_run_loop
    async def test_get_opening_timer(self):
        if self.test_device is None:
            self.skipTest("No RollerShutter device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        # Update its status
        await self.test_device.async_update()

        opening_timer = self.test_device.get_open_timer_duration_millis(channel=0)
        self.assertGreater(opening_timer, 0)

    @unittest_run_loop
    async def test_get_closing_timer(self):
        if self.test_device is None:
            self.skipTest("No RollerShutter device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        # Update its status
        await self.test_device.async_update()

        closing_timer = self.test_device.get_close_timer_duration_millis(channel=0)
        self.assertGreater(closing_timer, 0)

    @unittest_run_loop
    async def test_set_config(self):
        if self.test_device is None:
            self.skipTest("No RollerShutter device has been found to run this test on.")
        print(f"Testing device {self.test_device.name}")

        await self.test_device.async_update()

        # Retrieve original values
        original_open_timer = self.test_device.get_open_timer_duration_millis()
        original_close_timer = self.test_device.get_close_timer_duration_millis()

        # Set new random values
        open_timer = random.randint(10, 120)
        close_timer = random.randint(10, 120)
        await self.test_device.async_set_config(open_timer_seconds=open_timer, close_timer_seconds=close_timer,
                                                channel=0)

        opening_timer = self.test_device.get_open_timer_duration_millis(channel=0)
        self.assertEqual(opening_timer, open_timer * 1000)
        closing_timer = self.test_device.get_close_timer_duration_millis(channel=0)
        self.assertEqual(closing_timer, close_timer * 1000)

        # Restore original values
        await self.test_device.async_set_config(open_timer_seconds=int(original_open_timer/1000), close_timer_seconds=int(original_close_timer/1000))

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



import logging
from typing import Optional

from meross_iot.controller.device import BaseDevice, HubDevice, GenericSubDevice
from meross_iot.controller.mixins.consumption import ConsumptionXMixin, ConsumptionMixin
from meross_iot.controller.mixins.diffuser_light import DiffuserLightMixin
from meross_iot.controller.mixins.diffuser_spray import DiffuserSprayMixin
from meross_iot.controller.mixins.dnd import SystemDndMixin
from meross_iot.controller.mixins.electricity import ElectricityMixin
from meross_iot.controller.mixins.encryption import EncryptionSuiteMixin
from meross_iot.controller.mixins.garage import GarageOpenerMixin
from meross_iot.controller.mixins.hub import HubMts100Mixin, HubMixn, HubMs100Mixin
from meross_iot.controller.mixins.light import LightMixin
from meross_iot.controller.mixins.roller_shutter import RollerShutterTimerMixin
from meross_iot.controller.mixins.runtime import SystemRuntimeMixin
from meross_iot.controller.mixins.spray import SprayMixin
from meross_iot.controller.mixins.system import SystemAllMixin, SystemOnlineMixin
from meross_iot.controller.mixins.thermostat import ThermostatModeMixin, ThermostatModeBMixin
from meross_iot.controller.mixins.toggle import ToggleXMixin, ToggleMixin
from meross_iot.controller.subdevice import Mts100v3Valve, Ms100Sensor
from meross_iot.model.enums import Namespace
from meross_iot.model.exception import UnknownDeviceType
from meross_iot.model.http.device import HttpDeviceInfo
from meross_iot.model.http.subdevice import HttpSubdeviceInfo

_LOGGER = logging.getLogger(__name__)

_KNOWN_DEV_TYPES_CLASSES = {
    "mts100v3": Mts100v3Valve,
    "ms100": Ms100Sensor,
    "ms100f": Ms100Sensor
}

_ABILITY_MATRIX = {
    # Power plugs abilities
    Namespace.CONTROL_TOGGLEX.value: ToggleXMixin,
    Namespace.CONTROL_TOGGLE.value: ToggleMixin,
    Namespace.CONTROL_CONSUMPTIONX.value: ConsumptionXMixin,
    Namespace.CONTROL_CONSUMPTION.value: ConsumptionMixin,
    Namespace.CONTROL_ELECTRICITY.value: ElectricityMixin,

    # Encryption
    Namespace.SYSTEM_ENCRYPTION.value: EncryptionSuiteMixin,

    # Light abilities
    Namespace.CONTROL_LIGHT.value: LightMixin,

    # Garage opener
    Namespace.GARAGE_DOOR_STATE.value: GarageOpenerMixin,

    # Roller shutter timer
    Namespace.ROLLER_SHUTTER_STATE.value: RollerShutterTimerMixin,

    # Spray
    Namespace.CONTROL_SPRAY.value: SprayMixin,

    # Oil diffuser
    Namespace.DIFFUSER_LIGHT.value: DiffuserLightMixin,
    Namespace.DIFFUSER_SPRAY.value: DiffuserSprayMixin,

    # System
    Namespace.SYSTEM_ALL.value: SystemAllMixin,
    Namespace.SYSTEM_ONLINE.value: SystemOnlineMixin,
    Namespace.SYSTEM_RUNTIME.value: SystemRuntimeMixin,

    # Hub
    Namespace.HUB_ONLINE.value: HubMixn,
    Namespace.HUB_TOGGLEX.value: HubMixn,

    Namespace.HUB_SENSOR_ALL.value: HubMs100Mixin,
    Namespace.HUB_SENSOR_ALERT.value: HubMs100Mixin,
    Namespace.HUB_SENSOR_TEMPHUM.value: HubMs100Mixin,

    Namespace.HUB_MTS100_ALL.value: HubMts100Mixin,
    Namespace.HUB_MTS100_MODE.value: HubMts100Mixin,
    Namespace.HUB_MTS100_TEMPERATURE.value: HubMts100Mixin,

    # DND
    Namespace.SYSTEM_DND_MODE.value: SystemDndMixin,

    # Thermostat
    Namespace.CONTROL_THERMOSTAT_MODE.value: ThermostatModeMixin,
    Namespace.CONTROL_THERMOSTAT_MODEB.value: ThermostatModeBMixin,

    # TODO: BIND, UNBIND, ONLINE, WIFI, ETC!
}

_SUBDEVICE_MAPPING = {
    "mts100v3": Mts100v3Valve,
    "ms100": Ms100Sensor,
    "ms100f": Ms100Sensor
}

_dynamic_types = {}


def _caclulate_device_type_name(device_type: str, hardware_version: str, firmware_version: str) -> str:
    """
    Calculates the name of the dynamic-type for a specific class of devices
    :param device_type:
    :param hardware_version:
    :param firmware_version:
    :return:
    """
    return f"{device_type}:{hardware_version}:{firmware_version}"


def _lookup_cached_type(device_type: str, hardware_version: str, firmware_version: str) -> Optional[type]:
    """
    Returns the cached dynamic type for the specific device, if any was already built for that one.
    :param device_type:
    :param hardware_version:
    :param firmware_version:
    :return:
    """
    lookup_string = _caclulate_device_type_name(device_type, hardware_version, firmware_version).strip(":")
    return _dynamic_types.get(lookup_string)


def _build_cached_type(type_string: str, device_abilities: dict, base_class: type) -> type:
    """
    Builds a python type (class) dynamically by looking at the device abilities. In this way, we are able to
    "plugin" feature/mixins even for unknown new devices, given that they report abilities we already implemented.
    :param type_string:
    :param device_abilities:
    :return:
    """
    # Build a specific type at runtime by mixing plugins on-demand
    mixin_classes = set()

    # Add plugins by abilities
    for key, val in device_abilities.items():
        # When a device exposes the same ability like Tooggle and ToogleX, prefer the X version by filtering
        # out the non-X version.
        clsx = None
        cls = _ABILITY_MATRIX.get(key)

        # Check if for this ability the device exposes the X version
        x_key = f"{key}X"
        x_version_ability = device_abilities.get(x_key)
        if x_version_ability is not None:
            clsx = _ABILITY_MATRIX.get(x_key)

        # Now, if we have both the clsx and the cls, prefer the clsx, otherwise go for the cls
        if clsx is not None:
            mixin_classes.add(clsx)
        elif cls is not None:
            mixin_classes.add(cls)

    # We must be careful when ordering the mixin and leaving the BaseMerossDevice as last class.
    # Messing up with that will cause MRO to not resolve inheritance correctly.
    mixin_classes = list(mixin_classes)
    mixin_classes.append(base_class)
    m = type(type_string, tuple(mixin_classes), {"_abilities_spec": device_abilities})
    return m


def build_meross_device_from_abilities(http_device_info: HttpDeviceInfo,
                                       device_abilities: dict,
                                       manager) -> BaseDevice:
    """
    Builds a managed meross device object given the specs reported by HTTP api and the abilities reported by the device
    itself.

    :param http_device_info:
    :param device_abilities:
    :param manager:
    :return:
    """
    # The current implementation of this library is based on the usage of pluggable Mixin classes on top of
    # a couple of base implementations.
    _LOGGER.debug(f"Building managed device for {http_device_info.dev_name} ({http_device_info.uuid}). "
                  f"Reported abilities: {device_abilities}")

    # Check if we already have cached type for that device kind.
    cached_type = _lookup_cached_type(http_device_info.device_type,
                                      http_device_info.hdware_version,
                                      http_device_info.fmware_version)
    if cached_type is None:
        _LOGGER.debug(f"Could not find any cached type for {http_device_info.device_type},"
                      f"{http_device_info.hdware_version},"
                      f"{http_device_info.fmware_version}. It will be generated.")
        device_type_name = _caclulate_device_type_name(http_device_info.device_type,
                                                       http_device_info.hdware_version,
                                                       http_device_info.fmware_version)

        # Let's now pick the base class where to attach all the mixin.
        # We basically offer two possible base implementations:
        # - BaseMerossDevice: suitable for all non-hub devices
        # - HubMerossDevice: to be used when dealing with Hubs.
        # Unfortunately, it's not clear how we should discriminate an hub from a non-hub.
        # The current implementation decides which base class to use by looking at the presence
        # of 'Appliance.Hub.SubdeviceList': if exposed, we assume the device is a fully featured hub.
        discriminating_abilities = [Namespace.HUB_SUBDEVICELIST.value]
        base_class = BaseDevice
        if any (da in device_abilities for da in discriminating_abilities):
            _LOGGER.warning(f"Device {http_device_info.dev_name} ({http_device_info.device_type}, "
                            f"uuid {http_device_info.uuid}) reported one ability of {discriminating_abilities}. "
                            f"Assuming this is a full-featured HUB.")
            base_class = HubDevice

        cached_type = _build_cached_type(type_string=device_type_name,
                                         device_abilities=device_abilities,
                                         base_class=base_class)
        _dynamic_types[device_type_name] = cached_type

    #component = cached_type(device_uuid=http_device_info.uuid, manager=manager, **http_device_info.to_dict())
    component = cached_type(device_uuid=http_device_info.uuid, manager=manager, http_device_info=http_device_info)
    return component


def build_meross_device_from_known_types(http_device_info: HttpDeviceInfo,
                                         manager) -> BaseDevice:
    """
    Builds a managed meross device object by guess its relative class based on the device type string.
    Note that this method is capable of building managed device wrappers only if the device type is
    reported within the _KNOWN_DEV_TYPES_CLASSES. If your device type is not known yet, you should rely on
    `build_meross_device_from_abilities()` instead.

    :param http_device_info:
    :param manager:
    :return:
    """
    _LOGGER.debug(f"Building managed device for {http_device_info.dev_name} ({http_device_info.uuid}) "
                  f"from static known types ")
    dev_type = http_device_info.device_type.lower()
    target_clazz = _KNOWN_DEV_TYPES_CLASSES.get(dev_type)

    if target_clazz is None:
        _LOGGER.debug("Could not find any known device class for device type (%s).", http_device_info.device_type)
        raise UnknownDeviceType()

    return target_clazz(device_uuid=http_device_info.uuid, manager=manager, **http_device_info.to_dict())


def build_meross_subdevice(http_subdevice_info: HttpSubdeviceInfo, hub_uuid: str, hub_reported_abilities: dict,
                           manager) -> GenericSubDevice:
    _LOGGER.debug(f"Building managed device for {http_subdevice_info.sub_device_name} "
                  f"({http_subdevice_info.sub_device_id}).")

    # Build the device in accordance with the device type
    subdevtype = _SUBDEVICE_MAPPING.get(http_subdevice_info.sub_device_type)
    if subdevtype is None:
        _LOGGER.warning(f"Could not find any specific subdevice class for type {http_subdevice_info.sub_device_type}."
                        f" Applying generic SubDevice class.")
        subdevtype = GenericSubDevice
    return subdevtype(hubdevice_uuid=hub_uuid,
                      subdevice_id=http_subdevice_info.sub_device_id,
                      manager=manager,
                      **http_subdevice_info.to_dict())



from random import randint

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from meross_iot.controller.subdevice import Mts100v3Valve
from meross_iot.manager import MerossManager
from meross_iot.model.enums import ThermostatV3Mode, OnlineStatus
from meross_iot.model.plugin.hub import BatteryInfo
from tests import async_get_client
import os


if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestValve(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        await self.meross_manager.async_device_discovery()
        self.test_devices = self.meross_manager.find_devices(device_class=Mts100v3Valve, online_status=OnlineStatus.ONLINE)

    @unittest_run_loop
    async def test_ambient_temperature(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        for dev in self.test_devices:
            print(f"Testing device {dev.name}")
            res = await dev.async_update()
            temperature = await dev.async_get_temperature()
            self.assertIsInstance(temperature, float)

    @unittest_run_loop
    async def test_presets(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        for dev in self.test_devices:
            print(f"Testing device {dev.name}")
            res = await dev.async_update()

            max_supported_temp = dev.max_supported_temperature
            min_supported_temp = dev.min_supported_temperature

            self.assertGreater(len(dev.get_supported_presets()), 0)
            for preset in dev.get_supported_presets():
                old_preset_temp = dev.get_preset_temperature(preset)
                self.assertIsNotNone(old_preset_temp)
                new_preset = randint(min_supported_temp, max_supported_temp)
                await dev.async_set_preset_temperature(preset=preset, temperature=new_preset)
                new_current_preset = dev.get_preset_temperature(preset)
                self.assertEqual(new_current_preset, new_preset)

    @unittest_run_loop
    async def test_mode(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        self.assertIsNotNone(dev.mode)
        modes = set(ThermostatV3Mode)
        modes.remove(dev.mode)
        modes = list(modes)
        index = randint(0, len(modes)-1)
        target_mode = modes[index]

        await dev.async_set_mode(mode=target_mode)
        self.assertEqual(target_mode, dev.mode)

    @unittest_run_loop
    async def test_onoff(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        for dev in self.test_devices:
            print(f"Testing device {dev.name}")
            res = await dev.async_update()

            if dev.is_on():
                await dev.async_turn_off()
                self.assertFalse(dev.is_on())
                await asyncio.sleep(1)
                await dev.async_turn_on()
                self.assertTrue(dev.is_on())

    @unittest_run_loop
    async def test_battery(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev = self.test_devices[0]
        print(f"Testing device {dev.name}")
        await dev.async_update()
        res = await dev.async_get_battery_life()

        self.assertIsInstance(res, BatteryInfo)
        self.assertGreater(res.remaining_charge, -1)

    @unittest_run_loop
    async def test_push_notification(self):
        if len(self.test_devices) < 1:
            self.skipTest("No valve device has been found to run this test.")

        dev1 = self.test_devices[0]
        print(f"Testing device {dev1.name}")
        # Turn it on
        await dev1.async_turn_on()

        # Create a new manager
        new_meross_client, requires_logout = await async_get_client()
        m = None
        try:
            # Retrieve the same device with another manager
            m = MerossManager(http_client=new_meross_client)
            await m.async_init()
            await m.async_device_discovery()
            devs = m.find_devices(internal_ids=(dev1.internal_id,))
            dev = devs[0]

            await dev.async_update()
            await dev1.async_update()

            # Set target temperature to a random state
            target = randint(dev.min_supported_temperature, dev.max_supported_temperature)
            print(f"TARGET = {target}...")
            await dev1.async_set_target_temperature(temperature=target)

            # The manager that issues the command would immediately update the local state, so we can check
            # its update as soon as the command is issued.
            dev1_target_temp = dev1.target_temperature
            print(f"DEV1 = {dev1_target_temp}...")
            self.assertEqual(dev1_target_temp, target)

            # Wait a bit: give time for the push notification to get received on the other manager...
            await asyncio.sleep(5)
            # Make sure the other manager has received the push notification event
            dev_target_temp = dev.target_temperature
            print(f"DEV = {dev_target_temp}...")
            self.assertEqual(dev_target_temp, target)

        finally:
            if m is not None:
                m.close()
            if requires_logout:
                await new_meross_client.async_logout()

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



from meross_iot import name
from importlib import metadata
from importlib.metadata import PackageNotFoundError


def current_version():
    try:
        return metadata.version(name)
    except PackageNotFoundError:
        return "0.0.0"



import os
from uuid import uuid4
from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from meross_iot.controller.device import BaseDevice
from meross_iot.controller.mixins.toggle import ToggleXMixin
from meross_iot.manager import MerossManager
from meross_iot.model.constants import DEFAULT_MQTT_HOST, DEFAULT_MQTT_PORT
from meross_iot.model.enums import OnlineStatus, Namespace
from meross_iot.model.exception import CommandTimeoutError
from tests import async_get_client


if os.name == 'nt':
    import asyncio

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
else:
    import asyncio


class TestError(AioHTTPTestCase):
    async def get_application(self):
        return web.Application()

    async def setUpAsync(self):
        # Wait some time before next test-burst
        await asyncio.sleep(10)
        self.meross_client, self.requires_logout = await async_get_client()

        # Look for a device to be used for this test
        self.meross_manager = MerossManager(http_client=self.meross_client)
        await self.meross_manager.async_init()
        devices = await self.meross_manager.async_device_discovery()

    @unittest_run_loop
    async def test_invalid_target_device(self):
        async def send_command_to_unknown_device():
            random_uuid = uuid4()
            return await self.meross_manager.async_execute_cmd(destination_device_uuid=str(random_uuid),
                                                               method='GET',
                                                               namespace=Namespace.SYSTEM_ALL,
                                                               payload={},
                                                               mqtt_hostname=DEFAULT_MQTT_HOST,
                                                               mqtt_port=DEFAULT_MQTT_PORT)

        with self.assertRaises(CommandTimeoutError):
            await send_command_to_unknown_device()

    @unittest_run_loop
    async def test_invalid_namespace(self):
        devs = self.meross_manager.find_devices(device_class=ToggleXMixin, online_status=OnlineStatus.ONLINE)
        if len(devs) < 1:
            self.skipTest("No available/online devices found to test. Skipping...")
        dev = devs[0]
        print(f"Testing device {dev.name}")
        async def send_invalid_command_to_device(dev: BaseDevice):
            res = await self.meross_manager.async_execute_cmd(destination_device_uuid=dev.uuid,
                                                              method='GET',
                                                              namespace=Namespace.HUB_MTS100_MODE,
                                                              payload={},
                                                              mqtt_hostname=dev.mqtt_host,
                                                              mqtt_port=dev.mqtt_port)
            return res

        with self.assertRaises(CommandTimeoutError):
            await send_invalid_command_to_device(dev=dev)

    @unittest_run_loop
    async def test_invalid_payload(self):
        devs = self.meross_manager.find_devices(device_class=ToggleXMixin, online_status=OnlineStatus.ONLINE)
        if len(devs) < 1:
            self.skipTest("No available/online devices found to test. Skipping...")
        dev = devs[0]
        print(f"Testing device {dev.name}")
        async def send_invalid_command_to_device(dev: BaseDevice):
            return await self.meross_manager.async_execute_cmd(destination_device_uuid=dev.uuid,
                                                               method='SET',
                                                               namespace=Namespace.HUB_MTS100_MODE,
                                                               payload={'temperature': 'bar'},
                                                               mqtt_hostname=dev.mqtt_host,
                                                               mqtt_port=dev.mqtt_port)

        with self.assertRaises(CommandTimeoutError):
            await send_invalid_command_to_device(dev=dev)

    async def tearDownAsync(self):
        if self.requires_logout:
            await self.meross_client.async_logout()
        self.meross_manager.close()

        # Give a change to asyncio clean everything up
        await asyncio.sleep(1)



from __future__ import annotations  # Needed for returning the same class type where using factory static builders

import base64
import hashlib
import json
import logging
import os
import platform
import random
import string
import time
import uuid
from datetime import datetime
from typing import Optional, List

from aiohttp import ClientSession

from meross_iot.model.credentials import MerossCloudCreds
# Appears to be used as a part of the signature algorithm as constant "salt" (kinda useless)
from meross_iot.model.http.device import HttpDeviceInfo
from meross_iot.model.http.error_codes import ErrorCodes
from meross_iot.model.http.exception import TooManyTokensException, TokenExpiredException, AuthenticatedPostException, \
    HttpApiError, BadLoginException, BadDomainException, MissingMFA, WrongMFA
from meross_iot.model.http.subdevice import HttpSubdeviceInfo
from meross_iot.utilities.misc import current_version
from meross_iot.utilities.stats import HttpStatsCounter

_LOGGER = logging.getLogger(__name__)

_SECRET = "23x17ahWarFH6w29"
_LOGIN_URL = "%s/v1/Auth/signIn"
_LOG_URL = "%s/v1/log/user"
_DEV_LIST = "%s/v1/Device/devList"
_HUB_DUBDEV_LIST = "%s/v1/Hub/getSubDevices"
_LOGOUT_URL = "%s/v1/Profile/logout"
_MODULE_VERSION = current_version()
_DEFAULT_UA_HEADER = f"MerossIOT/{_MODULE_VERSION}"
_DEFAULT_APP_TYPE = "MerossIOT"
_DEFAULT_LOG_IDENTIFIER = '%030x' % random.randrange(16 ** 30) + str(uuid.uuid4())


class MerossHttpClient(object):
    """
    Utility class for dealing with Meross HTTP API.
    This class simplifies the usage of the Meross HTTP API providing login, logout and device listing API.
    """

    def __init__(self,
                 cloud_credentials: MerossCloudCreds,
                 http_proxy: str = None,
                 ua_header: str = _DEFAULT_UA_HEADER,
                 app_type: str = _DEFAULT_APP_TYPE,
                 app_version: str = _MODULE_VERSION,
                 log_identifier: str = _DEFAULT_LOG_IDENTIFIER):
        """
        Http library client constructor
        :param cloud_credentials: Credentials object to use for authentication
        :param http_proxy: HTTP proxy to use. When none, no proxy will be used. Defaults to None.
        :param ua_header: User agent headers to send alongside the various requests. The default value is auto-generated by the library.
        :param app_type: String used to discriminate the app type. The default value is auto-generated by the library.
        :param app_version: String used to discriminate the app version. The default value is auto-generated by the library.
        :param log_identifier: String used to log the app client usage. The default value is auto-generated by the library.
        """
        self._cloud_creds = cloud_credentials
        self._enable_proxy = False
        self._proxy_type = None
        self._proxy_addr = None
        self._http_proxy = http_proxy
        self._ua_header = ua_header
        self._stats_counter = HttpStatsCounter()
        self._log_identifier = log_identifier
        self._app_type = app_type
        self._app_version = app_version
        self._api_url = cloud_credentials.domain


    @property
    def stats(self) -> HttpStatsCounter:
        return self._stats_counter

    @property
    def cloud_credentials(self) -> MerossCloudCreds:
        """
        Returns the meross credentials returned by the the HTTP Api
        :return:
        """
        return self._cloud_creds

    @classmethod
    async def async_from_user_password(cls,
                                       api_base_url: str,
                                       email: str,
                                       password: str,
                                       http_proxy: str = None,
                                       ua_header: str = _DEFAULT_UA_HEADER,
                                       app_type: str = _DEFAULT_APP_TYPE,
                                       app_version: str = _MODULE_VERSION,
                                       log_identifier: str = _DEFAULT_LOG_IDENTIFIER,
                                       auto_retry_on_bad_domain: bool=True,
                                       mfa_code: string = None,
                                       *args, **kwargs) -> MerossHttpClient:
        """
        Builds a MerossHttpClient using username/password combination.
        In any case, the login will generate a token, which might expire at any time.

        :param api_base_url: Https base endpoint to use for API calls. It should be one of "https://iotx-eu.meross.com", "https://iotx-ap.meross.com" or "https://iotx-us.meross.com", based on your public IP region.,
        :param email: Meross account email
        :param password: Meross account password
        :param http_proxy: Optional http proxy to use when issuing the requests
        :param app_type: App Type header parameter to use
        :param app_version:  App Version header parameter to use
        :param log_identifier: Log identifier to use
        :param auto_retry_on_bad_domain: when set, it enables auto-retry when BadDomain exception occurs.
        :param mfa_code: multi-factor authentication code (optional)

        :return: an instance of `MerossHttpClient`
        """
        _LOGGER.debug(f"Logging in with email: {email}, password: XXXXX")
        creds = await cls.async_login(email=email,
                                      password=password,
                                      api_base_url=api_base_url,
                                      http_proxy=http_proxy,
                                      ua_header=ua_header,
                                      app_type=app_type,
                                      app_version=app_version,
                                      log_identifier=log_identifier,
                                      auto_retry_on_bad_domain=auto_retry_on_bad_domain,
                                      mfa_code=mfa_code)
        # Call log
        await cls._async_log(creds=creds,
                             api_base_url=api_base_url,
                             http_proxy=http_proxy,
                             ua_header=ua_header,
                             app_type=app_type,
                             app_version=app_version,
                             log_identifier=log_identifier)

        _LOGGER.debug(f"Login successful!")
        return MerossHttpClient(cloud_credentials=creds,
                                http_proxy=http_proxy,
                                ua_header=ua_header,
                                app_type=app_type,
                                app_version=app_version,
                                log_identifier=log_identifier)

    @classmethod
    async def async_from_cloud_creds(cls,
                                     creds: MerossCloudCreds,
                                     http_proxy: str = None,
                                     ua_header: str = _DEFAULT_UA_HEADER,
                                     app_type: str = _DEFAULT_APP_TYPE,
                                     app_version: str = _MODULE_VERSION,
                                     log_identifier: str = _DEFAULT_LOG_IDENTIFIER,
                                     *args,
                                     **kwargs) -> MerossHttpClient:
        """
        Build a MerossIot API client using the provided cloud-credentials object.
        :param creds: CloudCredentials object to build the client from
        :param http_proxy: HTTP proxy to use. When none, no proxy will be used. Defaults to None.
        :param ua_header: User agent headers to send alongside the various requests. The default value is auto-generated by the library.
        :param app_type: String used to discriminate the app type. The default value is auto-generated by the library.
        :param app_version: String used to discriminate the app version. The default value is auto-generated by the library.
        :param log_identifier: String used to log the app client usage. The default value is auto-generated by the library.
        :return:
        """

        # Use _log method to verify the credentials
        verify_creds = await cls._async_log(creds=creds,
                                            api_base_url=creds.domain,
                                            http_proxy=http_proxy,
                                            ua_header=ua_header,
                                            app_type=app_type,
                                            app_version=app_version,
                                            log_identifier=log_identifier)

        return MerossHttpClient(cloud_credentials=creds,
                                http_proxy=http_proxy,
                                ua_header=ua_header,
                                app_type=app_type,
                                app_version=app_version,
                                log_identifier=log_identifier)

    @classmethod
    async def async_login(cls,
                          api_base_url: str,
                          email: str,
                          password: str,
                          creds_env_var_name: str = '__MEROSS_CREDS',
                          http_proxy: str = None,
                          ua_header: str = _DEFAULT_UA_HEADER,
                          app_type: str = _DEFAULT_APP_TYPE,
                          app_version: str = _MODULE_VERSION,
                          log_identifier: str = _DEFAULT_LOG_IDENTIFIER,
                          country_code: str = "us",
                          agree_to_terms: int = 0,
                          mfa_code: str = None,
                          stats_counter: HttpStatsCounter = None,
                          auto_retry_on_bad_domain: bool = True,
                          *args, **kwargs) -> MerossCloudCreds:
        """
        Performs the login against the Meross HTTP endpoint.
        This api returns a MerossCloudCreds object, which contains the API endpoint, the mqtt endpoint and the
        authorization token to use the various API/MQTT services.
        Be cautious when invoking this API: asking for too many tokens as the Meross HTTP API might refuse
        to issue more tokens. Instead, you should keep using the same issued token when possible, possibly
        storing it across sessions. When you are done using a specific token, be sure to invoke logout
        to invalidate it.

        :param api_base_url: Meross API base url.
        :param email: Meross account email
        :param password: Meross account password
        :param creds_env_var_name: If set, indicate which env variables stores such credentials
        :param http_proxy: Optional http proxy to use when to performing the request
        :param ua_header: User Agent header to use when issuing the HTTP request
        :param stats_counter: Stats counter object
        :param app_type: App Type header parameter to use
        :param app_version:  App Version header parameter to use
        :param country_code: (experimental) The country code you are connecting from
        :param agree_to_terms: (optional) if 1 means we agree, if 0 or omitted, means we do not agree to terms
        :param mfa_code: (optional) MFA code
        :param auto_retry_on_bad_domain: when set, tells the library to retry the login when BadDomain occurs, osing the right domain value returned by the initial operation.

        :return: a `MerossCloudCreds` object
        """
        # Calculate password's md5
        m = hashlib.md5()
        m.update(password.encode("utf8"))
        hashed_password = m.hexdigest()
        data = { "email": email,
                 "password": hashed_password,
                 "accountCountryCode": country_code,
                 "encryption": 1,
                 "agree": 1 if agree_to_terms else 0,
                 "mobileInfo": {
                     "deviceModel": platform.machine(),
                     "mobileOsVersion": platform.version(),
                     "mobileOs": platform.system(),
                     "uuid": log_identifier,
                     "carrier":""
                 }
             }

        # MFACode is optional
        if mfa_code is not None:
            data["mfaCode"] = mfa_code

        url = _LOGIN_URL % api_base_url
        try:
            response_data = await MerossHttpClient._async_authenticated_post(url=url,
                                                                             params_data=data,
                                                                             mask_params_in_log=True,
                                                                             http_proxy=http_proxy,
                                                                             ua_header=ua_header,
                                                                             app_type=app_type,
                                                                             app_version=app_version,
                                                                             stats_counter=stats_counter)
        except BadDomainException as e:
            _LOGGER.error(f"Login API redirected to different region: {e.api_domain}. Login will be re-attempted")
            if auto_retry_on_bad_domain:
                return await MerossHttpClient.async_login(email=email, password=password, creds_env_var_name=creds_env_var_name, api_base_url=e.api_domain,http_proxy=http_proxy,ua_header=ua_header,app_type=app_type,app_version=app_version,log_identifier=log_identifier,country_code=country_code,agree_to_terms=agree_to_terms,mfa_code=mfa_code,stats_counter=stats_counter, auto_retry_on_bad_domain=False, *args, **kwargs)
            else:
                _LOGGER.exception(f"Login failed against {api_base_url}")
                raise e
        except HttpApiError as e:
            if e.error_code == ErrorCodes.MFA_CODE_REQUIRED:
                raise MissingMFA() from e
            elif e.error_code == ErrorCodes.WRONG_MFA_CODE:
                raise WrongMFA() from e
            raise e

        _LOGGER.info(f"Login successful against {api_base_url}")

        creds = MerossCloudCreds(
            token=response_data["token"],
            key=response_data["key"],
            user_id=response_data["userid"],
            user_email=response_data["email"],
            domain=response_data["domain"],
            mqtt_domain=response_data["mqttDomain"],
            issued_on=datetime.utcnow()
        )
        if creds_env_var_name is not None:
            os.environ[creds_env_var_name] = base64.b64encode(creds.to_json().encode("utf8")).decode("utf8")
        return creds

    @classmethod
    async def _async_authenticated_post(cls,
                                        url: str,
                                        params_data: dict,
                                        cloud_creds: Optional[MerossCloudCreds] = None,
                                        mask_params_in_log: bool = False,
                                        http_proxy: str = None,
                                        app_type: str = _DEFAULT_APP_TYPE,
                                        app_version: str = _MODULE_VERSION,
                                        ua_header: str = _DEFAULT_UA_HEADER,
                                        stats_counter: HttpStatsCounter = None
                                        ) -> dict:
        nonce = _generate_nonce(16)
        timestamp_millis = int(round(time.time() * 1000))
        encoded_params = _encode_params(params_data)

        # Generate the md5-hash (called signature)
        m = hashlib.md5()
        datatosign = '%s%s%s%s' % (_SECRET, timestamp_millis, nonce, encoded_params)
        m.update(datatosign.encode("utf8"))
        md5hash = m.hexdigest()

        headers = {
            "AppVersion": app_version,
            "Authorization": "Basic" if cloud_creds is None else "Basic %s" % cloud_creds.token,
            "vender": "meross",
            "AppType": app_type,
            "AppLanguage": "EN",
            "User-Agent": ua_header,
            "Content-Type": "application/json"
        }

        payload = {
            'params': encoded_params,
            'sign': md5hash,
            'timestamp': timestamp_millis,
            'nonce': nonce
        }

        # Perform the request.
        headers_with_masked_authrorization = headers.copy()
        if 'Authorization' in headers_with_masked_authrorization:
            headers_with_masked_authrorization['Authorization'] = 'XXXX-MASKED-XXXX'

        if not url.lower().startswith("http://") and not url.lower().startswith("https://"):
            _LOGGER.warning("Missing HTTP/HTTPS schema from Base API url. Assuming it's HTTPS://...")
            url = f"https://{url}"

        log_payload = payload
        if mask_params_in_log:
            log_payload = payload.copy()
            log_payload['params'] = 'XXXX-MASKED-XXXX'

        _LOGGER.debug(f"Performing HTTP request against {url}, headers: {headers}, post data: {payload}")
        async with ClientSession() as session:
            async with session.post(url, json=payload, headers=headers, proxy=http_proxy) as response:
                _LOGGER.debug(f"Response Status Code: {response.status}")
                # Check if that is ok.
                if response.status != 200:
                    if stats_counter is not None:
                        stats_counter.notify_http_request(request_url=url,
                                                          method="post",
                                                          http_response_code=response.status,
                                                          api_response_code=None)
                    raise AuthenticatedPostException("Failed request to API. Response code: %s" % str(response.status))

                # Save returned value
                jsondata = await response.json()
                code = jsondata.get('apiStatus')

                error = None
                try:
                    error = ErrorCodes(code)
                except ValueError as e:
                    raise AuthenticatedPostException(f"Unknown/Unhandled response code received from API. "
                                                     f"Response was: {jsondata}")
                finally:
                    # Keep track of the stats
                    if stats_counter is not None:
                        stats_counter.notify_http_request(request_url=url,
                                                          method="post",
                                                          http_response_code=response.status,
                                                          api_response_code=ErrorCodes.CODE_GENERIC_ERROR if error is None else error)
                    if error is None:
                        _LOGGER.error(f"Could not parse error code {code}.")
                    elif error == ErrorCodes.CODE_NO_ERROR:
                        return jsondata.get("data")
                    elif error in (ErrorCodes.CODE_TOKEN_EXPIRED, ErrorCodes.CODE_TOKEN_ERROR):
                        raise TokenExpiredException("The provided token has expired")
                    elif error == ErrorCodes.CODE_TOO_MANY_TOKENS:
                        raise TooManyTokensException("You have issued too many tokens without logging out and your "
                                                     "account might have been temporarly disabled.")
                    elif error in [ErrorCodes.CODE_WRONG_CREDENTIALS, ErrorCodes.CODE_UNEXISTING_ACCOUNT]:
                        raise BadLoginException("Invalid username/Password combination")
                    elif error == ErrorCodes.CODE_REDIRECT_REGION:
                        api_domain = jsondata.get("data").get("domain")
                        mqtt_domain = jsondata.get("data").get("mqttDomain")
                        raise BadDomainException(f"Invalid URL/API Endpoint used. Use: {api_domain} instead.", api_domain, mqtt_domain)
                    else:
                        _LOGGER.error(f"Received non-ok API status code: {error.name}. "
                                      f"Failed request to API. Response was: {jsondata}")
                        raise HttpApiError(error)

    async def async_logout(self,
                           *args,
                           **kwargs):
        """
        Invalidates the credentials stored in this object.

        :return: API response data
        """
        _LOGGER.debug(f"Logging out. Invalidating cached credentials {self._cloud_creds}")
        url = _LOGOUT_URL % self._api_url
        result = await MerossHttpClient._async_authenticated_post(url=url,
                                                                  params_data={},
                                                                  cloud_creds=self._cloud_creds,
                                                                  http_proxy=self._http_proxy,
                                                                  ua_header=self._ua_header,
                                                                  app_type=self._app_type,
                                                                  app_version=self._app_version,
                                                                  stats_counter=self._stats_counter)
        self._cloud_creds = None
        _LOGGER.info("Logout succeeded.")
        return result

    async def async_invalidate_credentials(self, creds: MerossCloudCreds):
        """
        Class method used to invalidate credentials without logging in with a full MerossHttpClient.

        :param creds: `MerossCloudCredentials` as returned by `async_login()` or `async_from_user_password()`
        :return: API response data
        """
        url = _LOGOUT_URL % self._api_url
        _LOGGER.debug(f"Logging out. Invalidating cached credentials {creds}")
        result = await MerossHttpClient._async_authenticated_post(url=url,
                                                                  params_data={},
                                                                  cloud_creds=creds,
                                                                  http_proxy=self._http_proxy,
                                                                  ua_header=self._ua_header,
                                                                  app_type=self._app_type,
                                                                  app_version=self._app_version,
                                                                  stats_counter=self._stats_counter)
        return result

    @classmethod
    async def _async_log(cls,
                         creds: MerossCloudCreds,
                         http_proxy: str = None,
                         ua_header: str = _DEFAULT_UA_HEADER,
                         app_type: str = _DEFAULT_APP_TYPE,
                         app_version: str = _MODULE_VERSION,
                         log_identifier: str = _DEFAULT_LOG_IDENTIFIER,
                         stats_counter: HttpStatsCounter = None,
                         *args,
                         **kwargs) -> dict:
        """
        Executes the LOG HTTP api. So far, it's still unknown whether this is needed and what it does.
        Most probably it logs the device specification to the remote endpoint for stats.

        :return:
        """
        data = {
            "system":platform.system(),
            "vendor":"meross",
            "uuid": log_identifier,
            "extra":"",
            "model":platform.machine(),
            "version":platform.version()
        }

        url = _LOG_URL % creds.domain
        return await cls._async_authenticated_post(url=url,
                                                   params_data=data,
                                                   cloud_creds=creds,
                                                   http_proxy=http_proxy,
                                                   ua_header=ua_header,
                                                   app_type=app_type,
                                                   app_version=app_version,
                                                   stats_counter=stats_counter
                                                   )

    async def async_list_devices(self,
                                 *args,
                                 **kwargs) -> List[HttpDeviceInfo]:
        """
        Asks to the HTTP api to list the Meross device belonging to the given user account.

        :return: a list of `HttpDeviceInfo`
        """
        url = _DEV_LIST % self._api_url
        result = await MerossHttpClient._async_authenticated_post(url=url,
                                                                  params_data={},
                                                                  cloud_creds=self._cloud_creds,
                                                                  http_proxy=self._http_proxy,
                                                                  ua_header=self._ua_header,
                                                                  app_type=self._app_type,
                                                                  app_version=self._app_version,
                                                                  stats_counter=self._stats_counter)
        return [HttpDeviceInfo.from_dict(x) for x in result]

    async def async_list_hub_subdevices(self,
                                        hub_id: str,
                                        *args,
                                        **kwargs) -> List[HttpSubdeviceInfo]:
        """
        Returns the sub-devices associated to the given hub.

        :param hub_id: Meross native UUID of the HUB

        :return: a list of `HttpSubdeviceInfo`
        """
        url = _HUB_DUBDEV_LIST % self._api_url
        result = await MerossHttpClient._async_authenticated_post(url=url,
                                                                  params_data={"uuid": hub_id},
                                                                  cloud_creds=self._cloud_creds,
                                                                  http_proxy=self._http_proxy,
                                                                  ua_header=self._ua_header,
                                                                  app_type=self._app_type,
                                                                  app_version=self._app_version,
                                                                  stats_counter=self._stats_counter)
        return [HttpSubdeviceInfo.from_dict(x) for x in result]

    def set_http_proxy(self, proxy_url: str):
        # TODO: validate proxy schema
        self._http_proxy = proxy_url


def _encode_params(parameters: dict):
    jsonstring = json.dumps(parameters)
    return str(base64.b64encode(jsonstring.encode("utf8")), "utf8")


def _generate_nonce(length: int):
    return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length))


def unauthenticated_command_executor(func, *args, **kwargs):
    def cmd(*iargs, **ikwargs):
        import asyncio

        # Fix event loop for Windows env
        if os.name == 'nt':
            import asyncio
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        params = {k: v for (k, v) in iargs[0].__dict__.items() if k not in ("func",)}

        loop = asyncio.get_event_loop()
        data = loop.run_until_complete(func(**params))
        print(data)
        loop.stop()

    return cmd


def authenticated_command_executor(method, *args, **kwargs):
    def cmd(*iargs, **ikwargs):
        import asyncio
        import os

        async def execute(method, *args, **kwargs):
            b64creds = os.getenv("__MEROSS_CREDS", None)
            if b64creds is None:
                raise ValueError("This method requires __MEROSS_CREDS env variable set. "
                                 "Please invoke auth login first.")

            jsoncreds = base64.b64decode(b64creds)
            creds = MerossCloudCreds.from_json(jsoncreds)
            client = await MerossHttpClient.async_from_cloud_creds(creds)
            m = getattr(client, method)
            return await m(*args, **kwargs)

        # Fix event loop for Windows env
        if os.name == 'nt':
            import asyncio
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        params = {k: v for (k, v) in iargs[0].__dict__.items() if k not in ("func", "credentials")}

        loop = asyncio.get_event_loop()
        data = loop.run_until_complete(execute(method, **params))
        print(data)
        loop.stop()

    return cmd


def main():
    import sys
    from argparse import ArgumentParser
    # Root parser
    parser = ArgumentParser(
        prog="meross_api_cli",
        description="Meross HTTP API utility",
        epilog="Created by Alberto Geniola")
    subparsers = parser.add_subparsers()

    # auth parser
    parser_auth = subparsers.add_parser('auth')
    auth_subparsers = parser_auth.add_subparsers()

    # auth login
    auth_login = auth_subparsers.add_parser('login',
                                            description="Logs in into meross HTTP api and prints out the credentials")
    auth_login.add_argument('--api-base-url', type=str, required=True, help='Api base URL to login against. Should be one among: https://iotx-eu.meross.com, https://iotx-us.meross.com, https://iotx-ap.meross.com')
    auth_login.add_argument('--email', type=str, required=True, help='Account username/email')
    auth_login.add_argument('--password', type=str, required=True, help='Account password')
    auth_login.add_argument('--creds-env-var-name', type=str, required=False,
                            help='When set, tells the library to store the credentials (upon successful login) within the indicated env var name')
    auth_login.set_defaults(func=unauthenticated_command_executor(MerossHttpClient.async_login))

    # auth logout
    auth_login = auth_subparsers.add_parser('logout', description="Logs out by invalidating the current token")
    auth_login.set_defaults(func=authenticated_command_executor("async_logout"))

    # device parser
    parser_device = subparsers.add_parser('device')
    device_subparsers = parser_device.add_subparsers()

    # devlist
    dev_list = device_subparsers.add_parser('list', description="Lists meross devices")
    dev_list.set_defaults(func=authenticated_command_executor("async_list_devices"))

    # hub parser
    parser_hub = subparsers.add_parser('hub')
    parser_hub.add_argument("--hub_id", required=True, help="Hub uuid")
    hub_subparsers = parser_hub.add_subparsers()

    # getsubdevices
    subdev_list = hub_subparsers.add_parser('list_sub_devices', description="Lists subdevices for the given hub")
    subdev_list.set_defaults(func=authenticated_command_executor("async_list_hub_subdevices"))

    data = parser.parse_args(sys.argv[1:])
    data.func(data)


if __name__ == '__main__':
    main()



from meross_iot.model.enums import Namespace
from meross_iot.model.push.common import HardwareInfo, FirmwareInfo, TimeInfo
from meross_iot.model.push.generic import GenericPushNotification


class BindPushNotification(GenericPushNotification):
    def __init__(self, originating_device_uuid: str, raw_data: dict):
        super().__init__(namespace=Namespace.CONTROL_BIND,
                         originating_device_uuid=originating_device_uuid,
                         raw_data=raw_data)

    @property
    def time(self) -> TimeInfo:
        return TimeInfo.from_dict(self.raw_data.get("bind").get("time"))

    @property
    def hwinfo(self) -> HardwareInfo:
        return TimeInfo.from_dict(self.raw_data.get("bind").get("hardware"))

    @property
    def fwinfo(self) -> FirmwareInfo:
        return TimeInfo.from_dict(self.raw_data.get("bind").get("firmware"))

